use eframe::egui;
use egui::*;

use image::DynamicImage;

use rfd::FileDialog;

use board_builder::{ Board, CoordTransformer, read_board_from_file, write_board_to_file };

use std::path::Path;

fn main() {
    let native_options = eframe::NativeOptions::default();
    eframe::run_native("Board Builder", native_options, Box::new(|cc| Box::new(BoardBuilderApp::new(cc))));
}

#[derive(Default)]
struct BoardBuilderApp {
    board: Board,
    texture: Option<TextureHandle>,
    image: Option<image::DynamicImage>,
    edit_mode: EditMode,
}

enum EditMode {
    Nodes,
    Edges,
}
impl Default for EditMode {
    fn default() -> Self {
        EditMode::Nodes
    }
}

impl eframe::App for BoardBuilderApp {
    fn update(&mut self, ctx: &Context, _frame: &mut eframe::Frame) {
        TopBottomPanel::top("menubar").show(ctx, |ui| {
            menu::bar(ui, |ui| {
                ui.menu_button("File", |ui| {
                    if ui.button("New").clicked() {
                        self.board = Board::default();
                        self.texture = None;
                        self.image = None;
                        self.edit_mode = EditMode::Nodes;
                    }
                    if ui.button("Open...").clicked() {
                        if let Some(board_file) = FileDialog::new().pick_file() {
                            match read_board_from_file(&board_file) {
                                Ok((board, image)) => {
                                    self.board = board;
                                    match image {
                                        None => { self.image = None; self.texture = None }
                                        Some(image) => self.load_image(ctx, image),
                                    }
                                }
                                Err(_) => panic!("Could not open file!"),
                            }
                        }
                    }
                    if ui.button("Save As...").clicked() {
                        if let Some(board_file) = FileDialog::new().save_file() {
                            write_board_to_file(&self.board, self.image.as_ref(), &board_file).expect("Something went wrong saving!");
                        }
                    }
                    if ui.button("Open Image...").clicked() {
                        let image_file = FileDialog::new()
                            .add_filter("Image", &["png", "jpg", "jpeg", "gif", "webp", "bmp", "tiff"])
                            .pick_file();
                        if let Some(image_file) = image_file {
                            self.load_image_file(ctx, &image_file).unwrap();
                        }
                    }
                });
                ui.menu_button("Edit", |ui| {
                    if ui.button("Edit Nodes").clicked() {
                        self.edit_mode = EditMode::Nodes;
                    }
                    if ui.button("Edit Edges").clicked() {
                        self.edit_mode = EditMode::Edges;
                    }
                    if ui.button("Edit Labels...").clicked() {
                        todo!();
                    }
                })
            });
        });
        CentralPanel::default().show(ctx, |ui| {
            if let Some(texture) = self.texture.as_ref() {
                let size = ui.available_size();
                let (response, painter) = ui.allocate_painter(size, Sense::click());
                let image = widgets::Image::new(texture, size);
                image.paint_at(ui, response.rect);
                let view = View(response.rect);
                self.draw_board(&painter, view);
            }
        });
    }
}

impl BoardBuilderApp {
    fn new(cc: &eframe::CreationContext<'_>) -> Self {
        let mut style = (*cc.egui_ctx.style()).clone();
        let mut button = style::TextStyle::Button.resolve(&style);
        button.size = 20.0;
        style.text_styles.insert(style::TextStyle::Button, button);
        cc.egui_ctx.set_style(style);

        BoardBuilderApp::default()
    }
    fn load_image_file(&mut self, ctx: &Context, image_file: &Path) -> Result<(), image::ImageError> {
        let image = image::io::Reader::open(image_file)?.decode()?;
        self.load_image(ctx, image);
        Ok(())
    }

    fn load_image(&mut self, ctx: &Context, image: DynamicImage) {
        let egui_image = egui::ColorImage::from_rgba_unmultiplied(
            [image.width() as _, image.height() as _],
            image.to_rgba8().as_flat_samples().as_slice(),
        );

        self.image = Some(image);
        self.texture = Some(ctx.load_texture("board-image", egui_image));
    }

    fn draw_board(&self, painter: &Painter, view: View) {
        for node in self.board.nodes.values() {
            painter.text(
                view.from_coords(node.x, node.y),
                Align2::CENTER_CENTER,
                &node.name,
                FontId::proportional(16.0),
                Color32::BLACK,
            );
            let stroke = Stroke { width: 1.0, color: Color32::BLACK };
            for edge in &node.edges {
                let other_node = &self.board.nodes[edge];
                painter.line_segment(
                    [view.from_coords(node.x, node.y), view.from_coords(other_node.x, other_node.y)],
                    stroke,
                );
            }
        }
    }
}

struct View(Rect);
impl CoordTransformer<Pos2> for View {
    fn origin(&self) -> Pos2 { self.0.min }
    fn extremes(&self) -> Pos2 { self.0.max }
}
